#define CYLCOORD
#define CONVERT_CYL_TO_RECT
/*****************************************************************************
*
* Copyright (c) 2000 - 2006, The Regents of the University of California
* Produced at the Lawrence Livermore National Laboratory
* All rights reserved.
*
* This file is part of VisIt. For details, see http://www.llnl.gov/visit/. The
* full copyright notice is contained in the file COPYRIGHT located at the root
* of the VisIt distribution or at http://www.llnl.gov/visit/copyright.html.
*
* Redistribution  and  use  in  source  and  binary  forms,  with  or  without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of  source code must  retain the above  copyright notice,
*    this list of conditions and the disclaimer below.
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this  list of  conditions  and  the  disclaimer (as noted below)  in  the
*    documentation and/or materials provided with the distribution.
*  - Neither the name of the UC/LLNL nor  the names of its contributors may be
*    used to  endorse or  promote products derived from  this software without
*    specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT  HOLDERS AND CONTRIBUTORS "AS IS"
* AND ANY EXPRESS OR  IMPLIED WARRANTIES, INCLUDING,  BUT NOT  LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND  FITNESS FOR A PARTICULAR  PURPOSE
* ARE  DISCLAIMED.  IN  NO  EVENT  SHALL  THE  REGENTS  OF  THE  UNIVERSITY OF
* CALIFORNIA, THE U.S.  DEPARTMENT  OF  ENERGY OR CONTRIBUTORS BE  LIABLE  FOR
* ANY  DIRECT,  INDIRECT,  INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR CONSEQUENTIAL
* DAMAGES (INCLUDING, BUT NOT  LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR
* SERVICES; LOSS OF  USE, DATA, OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER
* CAUSED  AND  ON  ANY  THEORY  OF  LIABILITY,  WHETHER  IN  CONTRACT,  STRICT
* LIABILITY, OR TORT  (INCLUDING NEGLIGENCE OR OTHERWISE)  ARISING IN ANY  WAY
* OUT OF THE  USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
*
*****************************************************************************/

// ************************************************************************* //
//                            avtIncite4HDF5ReaderFileFormat.C                           //
// ************************************************************************* //

#include <avtIncite4HDF5ReaderFileFormat.h>

#include <list>
#include <string>

#include <vtkFloatArray.h>
#include <vtkRectilinearGrid.h>
#include <vtkStructuredGrid.h>
#include <vtkUnstructuredGrid.h>

#include <avtDatabaseMetaData.h>

#include <Expression.h>

#include <InvalidVariableException.h>


using     std::string;

namespace avtIncite4HDF5ReaderAux {
  bool readDatasetDimensions(H5::H5File& h5File, int dims[3])
  {
    try {
      H5::Attribute cR = h5File.openGroup("/Coordinates").openAttribute("CoordinateRank");
      if (cR.getSpace().getSimpleExtentNdims() != 1) {
	std::cerr << "Error: Unexpected rank!" << std::endl;
	return false;
      }
      hsize_t nE;
      cR.getSpace().getSimpleExtentDims(&nE);
      if (nE != 3) {
	std::cerr << "Error: Unexpected number of dimensions!" << std::endl;
	return false;
      }
      cR.read(H5::PredType::NATIVE_INT, dims);
      return true;
    }
    catch (...) {
      return false;
    }

  }

  vtkFloatArray* readFloatDataSet(H5::H5File& h5File, const char* dataSetName, int expectedRank, hsize_t* dimsOut)
  {
    try {
      H5::DataSet dataset = h5File.openDataSet(dataSetName);
      H5::DataSpace dataspace = dataset.getSpace();

      if (dataspace.getSimpleExtentNdims() != expectedRank) {
	std::cerr << "Error: Unexpected rank!" << std::endl;
	return 0;
      }

      int ndims = dataspace.getSimpleExtentDims(dimsOut);

      hsize_t numElementsInDataSet = 1;
      for (int i=0; i<expectedRank; ++i) numElementsInDataSet *= dimsOut[i];

      vtkFloatArray *data = vtkFloatArray::New();
      data->SetNumberOfTuples(numElementsInDataSet);
      float *dataArray = (float *)data->GetVoidPointer(0);
      dataset.read(dataArray, H5::PredType::NATIVE_FLOAT);

      return data;
    }
    catch (...)
    {
      return 0;
    }
  }
}

// ****************************************************************************
//  Method: avtIncite4HDF5Reader constructor
//
//  Programmer: ghweber -- generated by xml2avt
//  Creation:   Mon Jan 29 16:44:46 PST 2007
//
// ****************************************************************************

  avtIncite4HDF5ReaderFileFormat::avtIncite4HDF5ReaderFileFormat(const char *filename)
: avtSTSDFileFormat(filename), mFile(filename, H5F_ACC_RDONLY)
{
  if (strstr(filename, "cyl") || strstr(filename, "Cyl") || strstr(filename, "CYL")) {
    mReadCylinderCoordinates = true;
  }
}


// ****************************************************************************
//  Method: avtIncite4HDF5ReaderFileFormat::FreeUpResources
//
//  Purpose:
//      When VisIt is done focusing on a particular timestep, it asks that
//      timestep to free up any resources (memory, file descriptors) that
//      it has associated with it.  This method is the mechanism for doing
//      that.
//
//  Programmer: ghweber -- generated by xml2avt
//  Creation:   Mon Jan 29 16:44:46 PST 2007
//
// ****************************************************************************

  void
avtIncite4HDF5ReaderFileFormat::FreeUpResources(void)
{
}


// ****************************************************************************
//  Method: avtIncite4HDF5ReaderFileFormat::PopulateDatabaseMetaData
//
//  Purpose:
//      This database meta-data object is like a table of contents for the
//      file.  By populating it, you are telling the rest of VisIt what
//      information it can request from you.
//
//  Programmer: ghweber -- generated by xml2avt
//  Creation:   Mon Jan 29 16:44:46 PST 2007
//
// ****************************************************************************

extern "C"  herr_t
add_var(hid_t loc_id, const char *varname, void *opData)
{
  std::list<std::string> *l = (std::list<std::string>*)opData;
  std::cout << "I: Adding variable " << varname << std::endl;
  l->push_back(varname);
}


  void
avtIncite4HDF5ReaderFileFormat::PopulateDatabaseMetaData(avtDatabaseMetaData *md)
{
  //
  // CODE TO ADD A MESH
  //
  string meshname = "mesh";
  //
  // AVT_RECTILINEAR_MESH, AVT_CURVILINEAR_MESH, AVT_UNSTRUCTURED_MESH,
  // AVT_POINT_MESH, AVT_SURFACE_MESH, AVT_UNKNOWN_MESH
  avtMeshType mt;
  if (mReadCylinderCoordinates)
#ifdef CONVERT_CYL_TO_RECT
    mt = AVT_CURVILINEAR_MESH;
#else
    mt = AVT_RECTILINEAR_MESH;
#endif
  else
    mt = AVT_CURVILINEAR_MESH;
  //
  int nblocks = 1;  //<-- this must be 1 for STSD
  int block_origin = 0;
  int spatial_dimension = 3;
  int topological_dimension = 3;
  double *extents = NULL;
  //
  // Here's the call that tells the meta-data object that we have a mesh:
  //
  AddMeshToMetaData(md, meshname, mt, extents, nblocks, block_origin,
      spatial_dimension, topological_dimension);

  std::list<std::string> varList;
  std::cout << "Scanning file for variables." << std::endl;
  herr_t idx = H5Giterate(mFile.getId(), "/DataValues", 0, add_var, &varList);
  for (std::list<std::string>::iterator it = varList.begin(); it!= varList.end(); ++it)
    AddScalarVarToMetaData(md, *it, "mesh", AVT_NODECENT);
  std::cout << "Done scanning file for variables." << std::endl;
  //
  // CODE TO ADD A VECTOR VARIABLE
  //
  // string mesh_for_this_var = meshname; // ??? -- could be multiple meshes
  // string varname = ...
  // int vector_dim = 2;
  //
  // AVT_NODECENT, AVT_ZONECENT, AVT_UNKNOWN_CENT
  // avtCentering cent = AVT_NODECENT;
  //
  //
  // Here's the call that tells the meta-data object that we have a var:
  //
  // AddVectorVarToMetaData(md, varname, mesh_for_this_var, cent,vector_dim);
  //

  //
  // CODE TO ADD A TENSOR VARIABLE
  //
  // string mesh_for_this_var = meshname; // ??? -- could be multiple meshes
  // string varname = ...
  // int tensor_dim = 9;
  //
  // AVT_NODECENT, AVT_ZONECENT, AVT_UNKNOWN_CENT
  // avtCentering cent = AVT_NODECENT;
  //
  //
  // Here's the call that tells the meta-data object that we have a var:
  //
  // AddTensorVarToMetaData(md, varname, mesh_for_this_var, cent,tensor_dim);
  //

  //
  // CODE TO ADD A MATERIAL
  //
  // string mesh_for_mat = meshname; // ??? -- could be multiple meshes
  // string matname = ...
  // int nmats = ...;
  // vector<string> mnames;
  // for (int i = 0 ; i < nmats ; i++)
  // {
  //     char str[32];
  //     sprintf(str, "mat%d", i);
  //     -- or -- 
  //     strcpy(str, "Aluminum");
  //     mnames.push_back(str);
  // }
  // 
  // Here's the call that tells the meta-data object that we have a mat:
  //
  // AddMaterialToMetaData(md, matname, mesh_for_mat, nmats, mnames);
  //
  //
  // Here's the way to add expressions:
  //Expression momentum_expr;
  //momentum_expr.SetName("momentum");
  //momentum_expr.SetDefinition("{u, v}");
  //momentum_expr.SetType(Expression::VectorMeshVar);
  //md->AddExpression(&momentum_expr);
  //Expression KineticEnergy_expr;
  //KineticEnergy_expr.SetName("KineticEnergy");
  //KineticEnergy_expr.SetDefinition("0.5*(momentum*momentum)/(rho*rho)");
  //KineticEnergy_expr.SetType(Expression::ScalarMeshVar);
  //md->AddExpression(&KineticEnergy_expr);
}


// ****************************************************************************
//  Method: avtIncite4HDF5ReaderFileFormat::GetMesh
//
//  Purpose:
//      Gets the mesh associated with this file.  The mesh is returned as a
//      derived type of vtkDataSet (ie vtkRectilinearGrid, vtkStructuredGrid,
//      vtkUnstructuredGrid, etc).
//
//  Arguments:
//      meshname    The name of the mesh of interest.  This can be ignored if
//                  there is only one mesh.
//
//  Programmer: ghweber -- generated by xml2avt
//  Creation:   Mon Jan 29 16:44:46 PST 2007
//
// ****************************************************************************

vtkDataSet * avtIncite4HDF5ReaderFileFormat::GetMesh(const char *meshname)
{
  std::cout << "I: In GetMesh()." << std::endl;
  if (!mReadCylinderCoordinates) {
    std::cout << "xyz" << std::endl;
    const int expectedRank = 1;
    int dims[3];
    if (!avtIncite4HDF5ReaderAux::readDatasetDimensions(mFile, dims)) return 0;
    hsize_t expectedNumElements = dims[0] * dims[1] * dims[2];

    try {
      H5::DataSet xDataset = mFile.openDataSet("/Coordinates/x");
      H5::DataSet yDataset = mFile.openDataSet("/Coordinates/y");
      H5::DataSet zDataset = mFile.openDataSet("/Coordinates/z");

      H5::DataSpace xDataspace = xDataset.getSpace();
      H5::DataSpace yDataspace = yDataset.getSpace();
      H5::DataSpace zDataspace = zDataset.getSpace();

      if (xDataspace.getSimpleExtentNdims() != expectedRank) {
	std::cerr << "Error: Unexpected rank for /Coordiantes/x!" << std::endl;
	return 0;
      }

      if (yDataspace.getSimpleExtentNdims() != expectedRank) {
	std::cerr << "Error: Unexpected rank for /Coordiantes/y!" << std::endl;
	return 0;
      }

      if (zDataspace.getSimpleExtentNdims() != expectedRank) {
	std::cerr << "Error: Unexpected rank for /Coordiantes/z!" << std::endl;
	return 0;
      }

      hsize_t xDims[expectedRank];
      hsize_t yDims[expectedRank];
      hsize_t zDims[expectedRank];
      xDataspace.getSimpleExtentDims(xDims);
      yDataspace.getSimpleExtentDims(yDims);
      zDataspace.getSimpleExtentDims(zDims);

      hsize_t numElementsInDataSet = 1;
      for (int i=0; i<expectedRank; ++i) {
	if (xDims[i] != yDims[i] || xDims[i] != zDims[i]) {
	  std::cerr << "Error: Dimensions of coordinate data sets do not match!" << std::endl;
	  return 0;
	}
	numElementsInDataSet *= xDims[i];
      }
      if (numElementsInDataSet != expectedNumElements) {
	std::cerr << "Error: Incorrect number of coordinates in data set." << std::endl;
	return 0;
      }

      float *xArray = new float[numElementsInDataSet];
      float *yArray = new float[numElementsInDataSet];
      float *zArray = new float[numElementsInDataSet];

      xDataset.read(xArray, H5::PredType::NATIVE_FLOAT);
      yDataset.read(yArray, H5::PredType::NATIVE_FLOAT);
      zDataset.read(zArray, H5::PredType::NATIVE_FLOAT);

      //
      // Create the vtkStructuredGrid and vtkPoints objects.
      //
      vtkStructuredGrid *sgrid  = vtkStructuredGrid::New();
      vtkPoints         *points = vtkPoints::New();
      sgrid->SetPoints(points);
      sgrid->SetDimensions(dims);
      std::cout << "I: Dimensions of mesh are " << dims[0] << " " << dims[1] << " " << dims[2] << std::endl;
      points->Delete();
      points->SetNumberOfPoints(numElementsInDataSet);

      //
      // Copy the coordinate values into the vtkPoints object.
      //
      float *pts = (float *) points->GetVoidPointer(0);
      float *xc = xArray;
      float *yc = yArray;
      float *zc = zArray;

      for(int k = 0; k < dims[2]; ++k) {
	for(int j = 0; j < dims[1]; ++j) {
	  for(int i = 0; i < dims[0]; ++i) {
	    *pts++ = *xc++;
	    *pts++ = *yc++;
	    *pts++ = *zc++;
	  }
	}
      }

      // Delete temporary arrays.
      delete [] xArray;
      delete [] yArray;
      delete [] zArray;

      return sgrid;
    }
    catch (...)
    {
      return 0;
    }
  }
  else {
    std::cout << "rthetaz" << std::endl;
    int ndims = 3;
    hsize_t dims[3] = {1,1,1};
    vtkFloatArray *coords[3] = {0,0,0};
    // Read the ndims and number of X,Y,Z nodes from file.

    coords[0] = avtIncite4HDF5ReaderAux::readFloatDataSet(mFile, "/Coordinates/r", 1, &dims[0]);
    coords[1] = avtIncite4HDF5ReaderAux::readFloatDataSet(mFile, "/Coordinates/theta", 1, &dims[1]);
    coords[2] = avtIncite4HDF5ReaderAux::readFloatDataSet(mFile, "/Coordinates/z", 1, &dims[2]);

#ifdef CONVERT_CYL_TO_RECT
    //
    // Create the vtkStructuredGrid and vtkPoints objects.
    //
    vtkStructuredGrid *sgrid  = vtkStructuredGrid::New();
    vtkPoints         *points = vtkPoints::New();
    sgrid->SetPoints(points);
    sgrid->SetDimensions(dims[0], dims[1], dims[2]);
    points->Delete();
    points->SetNumberOfPoints(dims[0] * dims[1] * dims[2]);

    //
    // Copy the coordinate values into the vtkPoints object.
    //
    float *pts = (float *) points->GetVoidPointer(0);

    for(int i = 0; i < dims[0]; ++i) {
      for(int j = 0; j < dims[1]; ++j) {
	float x = coords[0]->GetValue(i) * cos(coords[1]->GetValue(j));
	float y = coords[0]->GetValue(i) * sin(coords[1]->GetValue(j));
	for(int k = 0; k < dims[2]; ++k) {
	  pts[((k*dims[1]+j)*dims[0]+i)*3+0] = x;
	  pts[((k*dims[1]+j)*dims[0]+i)*3+1] = y;
	  pts[((k*dims[1]+j)*dims[0]+i)*3+2] = coords[2]->GetValue(k);
	}
      }
    }
    coords[0]->Delete();
    coords[1]->Delete();
    coords[2]->Delete();
    return sgrid;
#else
    //
    // Create the vtkRectilinearGrid object and set its dimensions
    // and coordinates.
    //
    vtkRectilinearGrid *rgrid = vtkRectilinearGrid::New();
    rgrid->SetDimensions(dims[0], dims[1], dims[2]);
    rgrid->SetXCoordinates(coords[0]);
    coords[0]->Delete();
    rgrid->SetYCoordinates(coords[1]);
    coords[1]->Delete();
    rgrid->SetZCoordinates(coords[2]);
    coords[2]->Delete();
    return rgrid;
#endif
  }
}

// ****************************************************************************
//  Method: avtIncite4HDF5ReaderFileFormat::GetVar
//
//  Purpose:
//      Gets a scalar variable associated with this file.  Although VTK has
//      support for many different types, the best bet is vtkFloatArray, since
//      that is supported everywhere through VisIt.
//
//  Arguments:
//      varname    The name of the variable requested.
//
//  Programmer: ghweber -- generated by xml2avt
//  Creation:   Mon Jan 29 16:44:46 PST 2007
//
// ****************************************************************************

  vtkDataArray *
avtIncite4HDF5ReaderFileFormat::GetVar(const char *varname)
{
  const char dsPrefix[] = "/DataValues/";
  char dsName[strlen(dsPrefix) + strlen(varname) + 1];
  strcpy(dsName, dsPrefix);
  strcat(dsName, varname);
  std::cout << "I: Reading values for variable " << varname << " from " << dsName << std::endl;
//#ifndef CYLCOORD
  hsize_t dim;
  return  avtIncite4HDF5ReaderAux::readFloatDataSet(mFile, dsName, 1, &dim);
//#else
//  hsize_t dims[3];
//  return  avtIncite4HDF5ReaderAux::readFloatDataSet(mFile, dsName, 3, dims);
//#endif

  //
  // If you have a file format where variables don't apply (for example a
  // strictly polygonal format like the STL (Stereo Lithography) format,
  // then uncomment the code below.
  //
  // EXCEPTION1(InvalidVariableException, varname);
  //

  //
  // If you do have a scalar variable, here is some code that may be helpful.
  //
  // int ntuples = XXX; // this is the number of entries in the variable.
  // vtkFloatArray *rv = vtkFloatArray::New();
  // rv->SetNumberOfTuples(ntuples);
  // for (int i = 0 ; i < ntuples ; i++)
  // {
  //      rv->SetTuple1(i, VAL);  // you must determine value for ith entry.
  // }
  //
  // return rv;
  //
}


// ****************************************************************************
//  Method: avtIncite4HDF5ReaderFileFormat::GetVectorVar
//
//  Purpose:
//      Gets a vector variable associated with this file.  Although VTK has
//      support for many different types, the best bet is vtkFloatArray, since
//      that is supported everywhere through VisIt.
//
//  Arguments:
//      varname    The name of the variable requested.
//
//  Programmer: ghweber -- generated by xml2avt
//  Creation:   Mon Jan 29 16:44:46 PST 2007
//
// ****************************************************************************

  vtkDataArray *
avtIncite4HDF5ReaderFileFormat::GetVectorVar(const char *varname)
{
  //
  // If you have a file format where variables don't apply (for example a
  // strictly polygonal format like the STL (Stereo Lithography) format,
  // then uncomment the code below.
  //
  // EXCEPTION1(InvalidVariableException, varname);

  //
  // If you do have a vector variable, here is some code that may be helpful.
  //
  // int ncomps = YYY;  // This is the rank of the vector - typically 2 or 3.
  // int ntuples = XXX; // this is the number of entries in the variable.
  // vtkFloatArray *rv = vtkFloatArray::New();
  // int ucomps = (ncomps == 2 ? 3 : ncomps);
  // rv->SetNumberOfComponents(ucomps);
  // rv->SetNumberOfTuples(ntuples);
  // float *one_entry = new float[ucomps];
  // for (int i = 0 ; i < ntuples ; i++)
  // {
  //      int j;
  //      for (j = 0 ; j < ncomps ; j++)
  //           one_entry[j] = ...
  //      for (j = ncomps ; j < ucomps ; j++)
  //           one_entry[j] = 0.;
  //      rv->SetTuple(i, one_entry); 
  // }
  //
  // delete [] one_entry;
  // return rv;
  //
}
